//-----------------------------------------------------------------------------
/*
Programmed by Olivier Sangala
http://olivier.sangala.free.fr
olivier_sangala@hotmail.com
*/
//-----------------------------------------------------------------------------

#define USE_TI89              // Compile for TI-89
#define USE_TI92PLUS          // Compile for TI-92 Plus
#define USE_V200              // Compile for V200

// #define OPTIMIZE_ROM_CALLS // Use ROM Call Optimization

#define MIN_AMS 100           // Compile for AMS 1.00 or higher

#define SAVE_SCREEN           // Save/Restore LCD Contents

#include <tigcclib.h>         // Include All Header Files

//---------------------------------------------------------------------------


#include <extgraph.h>

#include "collision.h"
#include "lines.h"
#include "motion.h"
#include "wf3d.h"

#include "main.h"

//---------------------------------------------------------------------------

extern int cos_table[256];
extern int sin_table[256];

//---------------------------------------------------------------------------

extern TObject *Camera;

extern int cutlines_enable;

//---------------------------------------------------------------------------

extern int clipping_enable;

//---------------------------------------------------------------------------

char * Plane0;

TGame*Game;

//---------------------------------------------------------------------------

// BLOCK #0 LIAISON #0 vertex

TPoint3D B0L0_Points[4]={
{-BLOCK_WIDTH/2, BLOCK_HEIGHT/2,0},
{ BLOCK_WIDTH/2, BLOCK_HEIGHT/2,0},
{ BLOCK_WIDTH/2,-BLOCK_HEIGHT/2,0},
{-BLOCK_WIDTH/2,-BLOCK_HEIGHT/2,0}
};

// SHIP vertex / lines

int ship_numpoints = 3;
TPoint3D VShip[3]={{0,0,0},{-4,0,-8},{4,0,-8},};

int ship_numlines = 3;
int LShip[3][2]={{0,1},{1,2},{2,0},};

//---------------------------------------------------------------------------


void NewGame(void)
{
	
	
	Tunnel_Init();
	Ship_Init();

	Game->ShipSpeed = 0;
	
	Game->CameraType = CT_BACKSKY;
	Game->ShipBlockID = 0;
	Game->show_type = (SHOW_LIAISONS|SHOW_BORDERS);
	
	Ship_OnNewAngle();
	
}
//---------------------------------------------------------------------------


void Tunnel_Init(void)
{
	
	TObject*O=&Game->OTunnel;
	
	// Position
	
	O->Pos=(TPoint3D){0,0,0};
	
	// Angle
	
	O->Angle=(TAngle3D){AO_XYZ,0,0,0};
	
	// add blocks
	
	Game->TunnelD.blocks_count=0;
	
	int a;
	for (a=0;a<BLOCK_MAX;a++)
	BlockAdd(0);
	
	
}
//---------------------------------------------------------------------------

void Ship_Init(void)
{
	
	
	int a;
	
	TObject*O=&Game->OShip;
	
	// Position
	
	O->Pos.x = 0;
	O->Pos.y = -BLOCK_HEIGHT/4;
	O->Pos.z = BLOCK_LENGTH/2;
	
	// Angle
	
	O->Angle=(TAngle3D){AO_ZYX,0,0,0};
	Ship_OnNewAngle();
	
	// Points
	
	for (a=0;a<ship_numpoints;a++)
	Game->VShip[a].P=VShip[a];
	
	
}
//---------------------------------------------------------------------------

void BlockAdd(char ay_diff)
{
	
	TLiaison *LiaisonPtr;
	
	// MOVE OLD BLOCKS
	
	TBlock*BPTR1=Game->TunnelD.B;
	TBlock*BPTR2=BPTR1+1;
	
	int b=0;
	for (b=0;b<BLOCK_MAX-1;b++,BPTR1++,BPTR2++)
	*BPTR1=*BPTR2;
	
	
	// BLOCKPTR
	
	TBlock*BlockPtr1=&Game->TunnelD.B[BLOCK_MAX-1];
	TBlock*BlockPtr0=&Game->TunnelD.B[BLOCK_MAX-2];
	
	// *********************
	// ***** LIAISON 0 *****
	// *********************
	
	// IF FIRST BLOCK
	
	if (Game->TunnelD.blocks_count==0)
	{
		BlockPtr1->ay=0;
		BlockPtr1->ay_diff=0;
		
		LiaisonPtr=&BlockPtr1->L[0];
		
		// SET B0L0
		int a;
		for (a=0;a<4;a++)
		LiaisonPtr->P[a].P=B0L0_Points[a];
		
		
	}
	
	// ELSE
	
	else
	{
		
		// FIT NEW BLOCK TO OLD BLOCK
		BlockPtr1->L[0]=BlockPtr0->L[2];
		BlockPtr1->ay=BlockPtr0->ay;
		
		BlockPtr1->ay_diff = ay_diff;
		
	}
	
	// *********************
	// ***** LIAISON 1 *****
	// *********************
	
	BlockPtr1->L[1]=BlockPtr1->L[0];
	LiaisonPtr=&BlockPtr1->L[1];
	
	// change angle
	BlockPtr1->ay+=BlockPtr1->ay_diff;
	
	// TURN LEFT
	if (BlockPtr1->ay_diff<0)
	{
		LiaisonPtr->P[1]=LiaisonPtr->P[0];
		P3D_MoveY(&LiaisonPtr->P[1].P,BLOCK_WIDTH,BlockPtr1->ay+64);
		
		LiaisonPtr->P[2]=LiaisonPtr->P[1];
		LiaisonPtr->P[2].P.y-=BLOCK_HEIGHT;
	}
	// TURN RIGHT
	else if (BlockPtr1->ay_diff>0)
	{
		LiaisonPtr->P[0]=LiaisonPtr->P[1];
		P3D_MoveY(&LiaisonPtr->P[0].P,BLOCK_WIDTH,BlockPtr1->ay-64);
		
		LiaisonPtr->P[3]=LiaisonPtr->P[0];
		LiaisonPtr->P[3].P.y-=BLOCK_HEIGHT;
	}
	
	
	// *********************
	// ***** LIAISON 2 *****
	// *********************
	
	BlockPtr1->L[2]=BlockPtr1->L[1];
	LiaisonPtr=&BlockPtr1->L[2];
	
	P3D_MoveY(&LiaisonPtr->P[0].P,BLOCK_LENGTH,BlockPtr1->ay);
	P3D_MoveY(&LiaisonPtr->P[1].P,BLOCK_LENGTH,BlockPtr1->ay);
	
	
	LiaisonPtr->P[2]=LiaisonPtr->P[1];
	LiaisonPtr->P[2].P.y-=BLOCK_HEIGHT;
	
	LiaisonPtr->P[3]=LiaisonPtr->P[0];
	LiaisonPtr->P[3].P.y-=BLOCK_HEIGHT;
	
	// ***********************
	// ***** BLOCK COUNT *****
	// ***********************
	
	if (Game->TunnelD.blocks_count<BLOCK_MAX)
	Game->TunnelD.blocks_count++;
	
	
	
}
//---------------------------------------------------------------------------

void BlockAddRandom(void)
{
	
	static int TurnCounter=0;
	static char ay_diff=0;
	
	
	if (!(TurnCounter&7))
	{
		//ay_diff=0;
		//while (!ay_diff)
		ay_diff=(random(8)<<2);
		
		if (random(2))
		ay_diff=-ay_diff;
	}
	
	BlockAdd(ay_diff);
	TurnCounter++;
	
	// translate all, to avoid coordinates overflow
	TranslateAll();
	
}
//---------------------------------------------------------------------------

int Block_IsPositionInside(TBlock*BlockPtr,TPoint3D*P)
{
	
	// Axis (Y)
	// not really useful
	
	if (abs(P->y)>(BLOCK_HEIGHT>>1))
	return 0;
	
	// Axis (X) & (Z)
	// uses 'PointLinePosition'
	
	TVertex*V0=&BlockPtr->L[0].P[0]; // DOWN - LEFT
	TVertex*V1=&BlockPtr->L[0].P[1]; // DOWN - RIGHT
	TVertex*V2=&BlockPtr->L[2].P[0]; // UP   - LEFT
	TVertex*V3=&BlockPtr->L[2].P[1]; // UP   - RIGHT
	
	// UP / DOWN
	
	int cd=PointLinePosition(V0->P.x,V0->P.z,V1->P.x,V1->P.z,P->x,P->z);
	int cu=PointLinePosition(V2->P.x,V2->P.z,V3->P.x,V3->P.z,P->x,P->z);
	
	if (((cd>0 && cu>0) || (cd<0 && cu<0)))
	return 0;
	
	// LEFT / RIGHT
	
	int cl=PointLinePosition(V0->P.x,V0->P.z,V2->P.x,V2->P.z,P->x,P->z);
	int cr=PointLinePosition(V1->P.x,V1->P.z,V3->P.x,V3->P.z,P->x,P->z);
	
	if (((cl>0 && cr>0) || (cl<0 && cr<0)))
	return 0;
	
	// ELSE
	
	return 1;
}
//---------------------------------------------------------------------------

void Ship_TickLCD(void)
{
	
	TObject*O=&Game->OShip;
	
	
	// move ship
	
	int a;
	for (a=0;a<Game->ShipSpeed;a++)
	{
		
		// PUSH POSITION
		
		TPoint3D P2=O->Pos;
		
		// move 1 pixel
		
		Game->MDShip.x=Game->OShip.Pos.z;
		Game->MDShip.y=Game->OShip.Pos.x;
		
		MD_MoveOnePixel(&Game->MDShip);
		
		Game->OShip.Pos.z=Game->MDShip.x;
		Game->OShip.Pos.x=Game->MDShip.y;
		
		
		
		// check collision
		// add block
		
		TBlock*BlockPtr0=&Game->TunnelD.B[0];
		TBlock*BlockPtr1=&Game->TunnelD.B[1];
		TBlock*BlockPtr2=&Game->TunnelD.B[2];
		
		int collision = 0;
		
		if (Game->ShipBlockID==0)
		{
			if (!Block_IsPositionInside(BlockPtr0,&O->Pos))
			{
				if (Block_IsPositionInside(BlockPtr1,&O->Pos))
				Game->ShipBlockID = 1;
				else
				collision = 1;
			}
		}
		else if (Game->ShipBlockID==1)
		{
			if (!Block_IsPositionInside(BlockPtr1,&O->Pos))
			{
				if (Block_IsPositionInside(BlockPtr0,&O->Pos))
				Game->ShipBlockID = 0;
				else if (Block_IsPositionInside(BlockPtr2,&O->Pos))
				{
					Game->ShipBlockID = 2;
					
					BlockAddRandom();
					Game->ShipBlockID--;
				}
				else
				collision = 1;
			}
		}
		
		
		// POP POSITION
		
		if (collision)
		{
			O->Pos=P2;
			break;
		}
		
		
	}
	
}
//---------------------------------------------------------------------------

void Ship_OnNewAngle(void)
{
	
	int dz = cos_table[Game->OShip.Angle.ay];
	int dx = sin_table[Game->OShip.Angle.ay];
	
	Game->MDShip.dx = dz;
	Game->MDShip.dy = dx;
	
	MD_OnNewSlope(&Game->MDShip);
	
}
//---------------------------------------------------------------------------

void TranslateAll(void)
{
	
	TLiaison *LP=&Game->TunnelD.B[0].L[0];
	
	// P2 = center of upper line of B0L0
	
	TPoint3D P2;
	
	P2.x = -((LP->P[0].P.x+LP->P[2].P.x)>>1);
	P2.y = 0;
	P2.z = -((LP->P[0].P.z+LP->P[2].P.z)>>1);
	
	// TUNNEL
	
	int b,l,p;
	
	for (b=0;b<BLOCK_MAX;b++)
	for (l=0;l<3;l++)
	for (p=0;p<4;p++)
	P3D_Translate(&Game->TunnelD.B[b].L[l].P[p].P,&P2);
	
	// SHIP
	
	P3D_Translate(&Game->OShip.Pos,&P2);
	
	// CAMERA (NOT USEFUL, AS ALWAYS REPLACED)
	
	//P3D_Translate(&Game->OCamera.Pos,&P2);
	
}
//---------------------------------------------------------------------------


void Camera_Place(void)
{
	
	
	TObject*O=Camera;
	TObject*O2=&Game->OShip;
	
	// camera
	// position : on ship
	// angle : looking at (Z)
	
	O->Pos = O2->Pos;
	O->Angle=(TAngle3D){AO_XYZ,0,0,0};
	
	
	if (Game->CameraType==CT_NORMAL)
	{
		// position : center of tunnel - behind ship
		O->Pos.y=0;
		P3D_MoveY(&O->Pos,(BLOCK_LENGTH>>1),O2->Angle.ay^128);
		
		O->Angle.AxisOrder=AO_YXZ;
		O->Angle.ay = O2->Angle.ay;
		
	}
	
	else if (Game->CameraType==CT_INTOSHIP)
	{
		O->Angle.AxisOrder=AO_ZYX;
		O->Angle.az = O2->Angle.az;
		O->Angle.ay = O2->Angle.ay;
	}
	
	else if (Game->CameraType==CT_SKY1)
	{
		// position : in the sky
		O->Pos.y+=64;
		
		O->Angle.AxisOrder=AO_XYZ;
		O->Angle.ax = -64;
		O->Angle.ay = O2->Angle.ay;
		
	}
	
	else if (Game->CameraType==CT_SKY2)
	{
		// position : in the sky
		O->Pos.y+=192;
		
		O->Angle.AxisOrder=AO_XYZ;
		O->Angle.ax = -64;
		O->Angle.ay = O2->Angle.ay;
		
	}
	
	else if (Game->CameraType==CT_SIDE)
	{
		// position : beside ship, in the sky
		O->Pos.y=0;
		P3D_MoveY(&O->Pos,BLOCK_LENGTH,O2->Angle.ay-64);
		O->Pos.y+=64;
		
		O->Angle.AxisOrder=AO_XYZ;
		O->Angle.ax = -32;
		O->Angle.ay = O2->Angle.ay+64;
		
	}
	
	else if (Game->CameraType==CT_BACKSKY)
	{
		// position : behind ship, in the sky
		O->Pos.y=0;
		P3D_MoveY(&O->Pos,BLOCK_LENGTH,O2->Angle.ay^128);
		O->Pos.y+=64;
		
		O->Angle.AxisOrder=AO_XYZ;
		O->Angle.ax = -32;
		O->Angle.ay = O2->Angle.ay;
		
	}
	
	
	
	
	
}
//---------------------------------------------------------------------------


void DrawScreen(void)
{
	
	Camera_Place();
	
	ClrScr();
	
	DrawTunnel();
	
	DrawShip();
	
	// DEBUG
	/*
	char str[128];
	
	sprintf(str," %d %d %d ***** %d %d %d ",
	Game->TunnelD.B[0].L[0].P[0].P3.x,
	Game->TunnelD.B[0].L[0].P[0].P3.y,
	Game->TunnelD.B[0].L[0].P[0].P3.z,
	Game->TunnelD.B[1].L[0].P[0].P3.x,
	Game->TunnelD.B[1].L[0].P[0].P3.y,
	Game->TunnelD.B[1].L[0].P[0].P3.z);
	
	DrawStr(0,0,str,A_NORMAL);
	*/
	
	FastCopyScreen(Plane0,LCD_MEM);
	
}
//---------------------------------------------------------------------------


void DrawTunnel(void)
{
	
	// CALC
	// L0 for each BLOCK
	
	int b,p;
	
	for (b=0;b<BLOCK_SHOWN+1;b++)
	for (p=0;p<4;p++)
	Vertex_Calc(&Game->TunnelD.B[b].L[0].P[p],&Game->OTunnel);
	
	// DRAW LINES
	
	for (b=0;b<BLOCK_SHOWN;b++)
	{
		TBlock *BlockPtr0=&Game->TunnelD.B[b];
		TBlock *BlockPtr1=&Game->TunnelD.B[b+1];
		
		// Liaisons
		
		if (Game->show_type&SHOW_LIAISONS)
		{
			TLiaison*LPTR=&BlockPtr0->L[0];
			
			Line_Draw(&LPTR->P[0],&LPTR->P[1]);
			Line_Draw(&LPTR->P[1],&LPTR->P[2]);
			Line_Draw(&LPTR->P[2],&LPTR->P[3]);
			Line_Draw(&LPTR->P[3],&LPTR->P[0]);
		}
		
		// Borders
		
		if (Game->show_type&SHOW_BORDERS)
		{
			for (p=0;p<4;p++)
			Line_Draw(&BlockPtr0->L[0].P[p],&BlockPtr1->L[0].P[p]);
		}
		
	}
	
}
//---------------------------------------------------------------------------

void DrawShip(void)
{
	
	// CALC
	
	int a;
	for (a=0;a<ship_numpoints;a++)
	Vertex_Calc(&Game->VShip[a],&Game->OShip);
	
	// DRAW LINES
	
	for (a=0;a<ship_numlines;a++)
	Line_Draw(&Game->VShip[LShip[a][0]],&Game->VShip[LShip[a][1]]);
	
	
	
}
//---------------------------------------------------------------------------

// Main Function
void _main(void)
{
	
	if (!(Plane0=malloc(LCD_SIZE)))
	return;
	
	
	TGame GameA;
	Game=&GameA;
	
	Camera=&Game->OCamera;
	
	clrscr();
	FontSetSys(0);
	printf("Tunnel3D 1.4 \n");
	printf("Programmed by Olivier Sangala \n");
	printf("http://olivier.sangala.free.fr \n");
	printf("olivier_sangala@hotmail.com \n");
	FontSetSys(1);
	ngetchx();
	
	FontSetSys(0);
	randomize();
	
	
	NewGame();
	
	PortSet(Plane0,239,127);
	
	// keyboard options
	int osikid = OSInitKeyInitDelay (10);
	int osibkd = OSInitBetweenKeyDelay (10);
	
	for (;;)
	{
		
		
		Ship_TickLCD();
	
		DrawScreen();
		
		if (!kbhit())
		continue;
		
		int key = ngetchx();
		
		if (key==KEY_LEFT)
		{
			Game->OShip.Angle.ay-=4;
			Ship_OnNewAngle();
		}
		else if (key==KEY_RIGHT)
		{
			Game->OShip.Angle.ay+=4;
			Ship_OnNewAngle();
		}
		
		else if (key==KEY_UP)
		{
			Game->ShipSpeed++;
			if (Game->ShipSpeed>SPEED_MAX)
			Game->ShipSpeed=SPEED_MAX;
		}
		else if (key==KEY_DOWN)
		{
			Game->ShipSpeed--;
			if (Game->ShipSpeed<0)
			Game->ShipSpeed=0;
		}
		
		
		else if (key==KEY_F1)
		{
			Game->CameraType++;
			Game->CameraType%=CT_MAX;
		}
		
		else if (key==KEY_F2)
		{
			Game->show_type++;
			Game->show_type&=3;
		}
		
		else if (key==KEY_F3)
		clipping_enable^=1;
		else if (key==KEY_F4)
		cutlines_enable^=1;
		
		
		
		else if (key==KEY_ESC)
		break;
		
	}
	
	PortRestore();
	
	// keyboard options
	OSInitKeyInitDelay (osikid);
	OSInitBetweenKeyDelay (osibkd);
	
	if (Plane0)
	free(Plane0);
	
}
