//-----------------------------------------------------------------------------
/*
Programmed by Olivier Sangala
http://olivier.sangala.free.fr
olivier_sangala@hotmail.com
*/
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// WIREFRAME 3D FUNCTIONS
//-----------------------------------------------------------------------------

/*

Direction of axes :

(X) : LEFT -> RIGHT
(Y) : BOTTOM -> TOP
(Z) : BACKWARD -> FORWARD

Direction of angles :

(ax) : (Z)->(Y)
(ay) : (Z)->(X)
(az) : (Y)->(X)

Default angle (camera) :
looking at (Z)

Angles : 

0-360 -> 0-256

*/

//-----------------------------------------------------------------------------

enum {AO_XYZ,AO_XZY,AO_YXZ,AO_YZX,AO_ZXY,AO_ZYX};

//---------------------------------------------------------------------------

typedef unsigned char byte;

typedef struct
{
	int x,y;
}
TPoint2D;

typedef struct
{
	int x,y,z;
}
TPoint3D;

typedef struct
{
	TPoint3D P,P2,P3;
	TPoint2D P4;
}
TVertex;

typedef struct
{
	int AxisOrder;
	byte ax,ay,az;
}
TAngle3D;

typedef struct
{
	TPoint3D Pos;
	TAngle3D Angle;
}
TObject;

//---------------------------------------------------------------------------

void Calc_RX(TPoint3D *P1, TObject *O);
void Calc_RY(TPoint3D *P1, TObject *O);
void Calc_RZ(TPoint3D *P1, TObject *O);

void Object_InvertAngle(TObject *O);

void P3D_Rotate(TPoint3D *P1, TObject *O, TPoint3D *P2);
void P3D_CalcR2A(TPoint3D *P1, TObject *O, TPoint3D *P2);
void P3D_CalcA2R(TPoint3D *P1, TObject *O, TPoint3D *P2);
void P3D_Project(TPoint3D *P1, TPoint2D *P2);
void P3D_Translate(TPoint3D *P1, TPoint3D *P2);
void P3D_MoveY(TPoint3D *P1,int Radius, byte ay);

void Vertex_Calc(TVertex *V, TObject *O);

void Line_Draw(TVertex * V1, TVertex * V2);

//---------------------------------------------------------------------------

